#-----------------------------------------------------------------------------
#           Name: ogl_cg_multitexture.cpp
#         Author: Kevin Harris
#  Last Modified: 04/26/05
#    Description: This sample demonstrates how to blend two textures together
#                 with Cg using either OpenGL's native multi-texture support 
#                 (using semantics) or by using Cg's special texture functions:
#                 cgGLSetTextureParameter, cgGLEnableTextureParameter, and
#                 cgGLDisableTextureParameter.
#-----------------------------------------------------------------------------
#
# Original C++ code by Kevin Harris (kevin@codesampler.com)
# See www.codesampler.com for the original files
# OpenGL samples page 10: Multi-Texturing with Cg
#
# Modified for Tcl3D by Paul Obermeier 2007/05/22
# See www.tcl3d.org for the Tcl3D extension.
#
# The original demo has been extended with a little GUI to allow switching
# between the two call semantics at runtime.
# To visualize, that a different shader program is active,
# the OpenGL semantics shader adds only half of the checker image color.

package require Tk
package require Img
package require tcl3d 0.4.0

set g_winWidth  640
set g_winHeight 480

set g_LastMousePosX(1) 0
set g_LastMousePosY(1) 0
set g_fSpinX(1) 0.0
set g_fSpinY(1) 0.0

set g_testTextureInd    0
set g_checkerTextureInd 1

set g_callSemantics "CG"
set g_isProgramLoaded false

# Determine the directory of this script.
set g_scriptDir [file dirname [info script]]

#       GL_T2F_C3F_V3F
#       tu  tv   r   g   b     x    y    z 
set g_quadVertices [tcl3dVectorFromArgs GLfloat \
        0.0 0.0  1.0 1.0 0.0  -1.0 -1.0  0.0 \
        1.0 0.0  1.0 0.0 0.0   1.0 -1.0  0.0 \
        1.0 1.0  0.0 1.0 0.0   1.0  1.0  0.0 \
        0.0 1.0  0.0 0.0 1.0  -1.0  1.0  0.0]

# Show errors occuring in the Togl callbacks.
proc bgerror { msg } {
    tk_messageBox -icon error -type ok -message "Error: $msg\n\n$::errorInfo"
    exit
}

# Print OS and GL message into widget a the bottom of the window.
proc PrintInfo { msg } {
    if { [winfo exists .fr.info] } {
        .fr.info configure -text $msg
    }
}

# Print actual shader info message into widget a the bottom of the window.
proc PrintShaderInfo {} {
    if { [winfo exists .fr.shaderinfo] } {
        if { $::g_isProgramLoaded } {
            set profileStr [tcl3dCgFindProfileByNum $::g_CGprofile]
            set msg [format "Profile: %s File: %s" $profileStr $::g_shaderFile]
        } else {
            set msg "No Cg program loaded"
        }
        .fr.shaderinfo configure -text $msg
    }
}

proc SetMouseInput { x y } {
    set ::g_LastMousePosX(1) $x
    set ::g_LastMousePosY(1) $y
}

proc GetMouseInput { x y } {
    set nXDiff [expr ($x - $::g_LastMousePosX(1))]
    set nYDiff [expr ($y - $::g_LastMousePosY(1))]
        
    set ::g_fSpinX(1) [expr $::g_fSpinX(1) - $nXDiff]
    set ::g_fSpinY(1) [expr $::g_fSpinY(1) - $nYDiff]

    set ::g_LastMousePosX(1) $x
    set ::g_LastMousePosY(1) $y
    .fr.toglwin postredisplay
}

# Get the full path name of the shader file according to the call semantics
# currently set. Take into account, that this program may be running from 
# within a Starpack.

proc GetShaderFileName {} {
    if { $::g_callSemantics eq "CG" } {
        set ::g_shaderFile "ogl_cg_multitexture-cg.cg"
    } else {
        set ::g_shaderFile "ogl_cg_multitexture-gl.cg"
    }
    set cgFile [tcl3dGetExtFile [file join $::g_scriptDir $::g_shaderFile]]
    return $cgFile
}

proc LoadFileIntoTextWidget { w fileName } {
    set retVal [catch {open $fileName r} fp]
    if { $retVal != 0 } {
        error "Could not open file $fileName for reading."
    }
    $w delete 1.0 end
    while { ![eof $fp] } {
        $w insert end [read $fp 2048]
    }
    close $fp
}

proc TextWidgetToFile { w fileName } {
    if { [info exists      ::starkit::topdir] && \
         [file isdirectory $::starkit::topdir] } {
        set dumpDir [file dirname $::starkit::topdir]
    } else {
        set dumpDir [file dirname $fileName]
    }
    set dumpFile [file tail $fileName]

    set saveName [file join $dumpDir $dumpFile]
    set retVal [catch {open $saveName w} fp]
    if { $retVal != 0 } {
        error "Could not open file $saveName for writing."
    }
    puts -nonewline $fp [$w get 1.0 end]
    close $fp
}

proc AddMenuCmd { menu label acc cmd } {
    $menu add command -label $label -accelerator $acc -command $cmd
}

proc ShowEditor { fileName { textStr "" } } {
    global gPres

    # Font to be used in the text widget.
    set textFont {-family {Courier} -size 10}

    if { ! [info exists gPres(editCount)] } {
        set gPres(editCount) 0
    } else {
        incr gPres(editCount)
    }

    set titleStr "[file tail $fileName]"
    set tw ".poTextEdit_$gPres(editCount)"

    toplevel $tw
    wm title $tw $titleStr
    frame $tw.workfr -relief sunken -borderwidth 1
    pack  $tw.workfr -side top -fill both -expand 1

    set hMenu $tw.menufr
    menu $hMenu -borderwidth 2 -relief sunken
    $hMenu add cascade -menu $hMenu.file -label File -underline 0

    set textId [tcl3dCreateScrolledText $tw.workfr "$fileName" -font $textFont]

    set fileMenu $hMenu.file
    menu $fileMenu -tearoff 0
    AddMenuCmd $fileMenu "Save" "Ctrl+S" "TextWidgetToFile $textId [list $fileName]"
    AddMenuCmd $fileMenu "Close" "Ctrl+W" "destroy $tw"

    bind $tw <Control-s> "TextWidgetToFile $textId [list $fileName]"
    bind $tw <Control-w> "destroy $tw"
    bind $tw <Escape>    "destroy $tw"
    wm protocol $tw WM_DELETE_WINDOW "destroy $tw"

    $tw configure -menu $hMenu 

    if { $textStr eq "" } {
        LoadFileIntoTextWidget $textId $fileName
    } else {
        $textId delete 1.0 end
        $textId insert end $textStr
    }
    focus $tw
}

proc ViewShaderFile {} {
    ShowEditor [GetShaderFileName]
}

# Toggle between the supported call semantics. The current semantics is
# stored in global variable g_callSemantics, which itself is bound to
# radiobuttons on the GUI.

proc ToggleSemantics { toglwin } {
    CleanupShader
    set ::g_isProgramLoaded [InitShader]
    PrintShaderInfo
    $toglwin postredisplay
}

proc LoadTexture { name index } {
    set texName [file join $::g_scriptDir $name]
    set retVal [catch {set phImg [image create photo -file $texName]} err1]
    if { $retVal != 0 } {
        error "Error reading image $texName ($err1)"
    } else {
        set w [image width  $phImg]
        set h [image height $phImg]
        set n [tcl3dPhotoChans $phImg]
        set pTexImage [tcl3dVectorFromPhoto $phImg]
        image delete $phImg
    }

    set ::g_textureID($index) [tcl3dVector GLuint 1]
    glGenTextures 1 $::g_textureID($index)

    glBindTexture GL_TEXTURE_2D [$::g_textureID($index) get 0]

    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MIN_FILTER $::GL_LINEAR
    glTexParameteri GL_TEXTURE_2D GL_TEXTURE_MAG_FILTER $::GL_LINEAR

    glTexImage2D GL_TEXTURE_2D 0 3 $w $h 0 \
                 $::GL_RGBA GL_UNSIGNED_BYTE $pTexImage
    $pTexImage delete
}

proc InitShader {} {
    # Search for a valid pixel shader profile in this order:
    #
    # CG_PROFILE_ARBFP1 - GL_ARB_fragment_program
    # CG_PROFILE_FP30   - GL_NV_fragment_program
    # CG_PROFILE_FP20   - NV_texture_shader & NV_register_combiners
    
    set ::g_CGprofile [tcl3dCgFindProfile $::CG_PROFILE_ARBFP1 \
                                          $::CG_PROFILE_FP30 \
                                          $::CG_PROFILE_FP20]
    if { $::g_CGprofile eq "" } {
        tk_messageBox -icon error -type ok \
                      -title "tcl3dCgFindProfile" \
                      -message "Failed to initialize fragment shader.
                               No ARBFP1, FP30 or FP20 profile available."
        return false
    }

    # Create a CG context
    set ::g_CGcontext [cgCreateContext]

    # Create the pixel shader...
    set cgFile [GetShaderFileName]
    set ::g_CGprogram [cgCreateProgramFromFile $::g_CGcontext \
                                               CG_SOURCE $cgFile \
                                               $::g_CGprofile \
                                               "main" "NULL"]
    set retVal [tcl3dCgGetError $::g_CGcontext]
    if { $retVal ne "" } {
        tk_messageBox -icon error -type ok \
                      -title "cgCreateProgramFromFile" \
                      -message "$retVal"
        return false
    }

    # Load the programs using Cg's expanded interface...
    cgGLLoadProgram $::g_CGprogram
    set retVal [tcl3dCgGetError $::g_CGcontext]
    if { $retVal ne "" } {
        tk_messageBox -icon error -type ok \
                      -title "cgGLLoadProgram" \
                      -message "$retVal"
        return false
    }

    # Bind some parameters by name so we can set them later...
    if { $::g_callSemantics eq "CG" } {
        set ::g_CGparam_testTexture \
            [cgGetNamedParameter $::g_CGprogram "testTexture"]
        set ::g_CGparam_checkerTexture \
            [cgGetNamedParameter $::g_CGprogram "checkerTexture"]
    }
    return true
}

proc checkExtProc { extProc } {
     if { ![tcl3dOglHaveFunc $extProc] } {
        error "Extension proc $extProc not available"
    }
}

proc tclCreateFunc { toglwin } {
    # Needed to correctly initialize the OpenGL extension library.
    tcl3dOglExtInit

    glEnable GL_TEXTURE_2D
    glEnable GL_DEPTH_TEST

    glClearColor 0.0 0.0 0.0 1.0

    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 45.0 [expr double($::g_winWidth)/double($::g_winHeight)] \
                   0.1 100.0

    # Check for GL_ARB_multitexture extension.
    if { ! [tcl3dOglHaveExtension "GL_ARB_multitexture"] } {
        error "Extension GL_ARB_multitexture missing"
    }
    checkExtProc "glActiveTextureARB"
    checkExtProc "glMultiTexCoord2fARB"
    checkExtProc "glClientActiveTextureARB"

    LoadTexture "test.bmp"    $::g_testTextureInd
    LoadTexture "checker.bmp" $::g_checkerTextureInd

    set ::g_isProgramLoaded [InitShader]
}

proc tclReshapeFunc { toglwin w h } {
    set ::g_WinWidth  $w
    set ::g_WinHeight $h
    glViewport 0 0 $w $h
    glMatrixMode GL_PROJECTION
    glLoadIdentity
    gluPerspective 45.0 [expr double($w)/double($h)] 0.1 100.0
}

proc tclDisplayFunc { toglwin } {
    glClear [expr $::GL_COLOR_BUFFER_BIT | $::GL_DEPTH_BUFFER_BIT]

    # Viewport command is not really needed, but has been inserted for
    # Mac OSX. Presentation framework (Tk) does not send a reshape event,
    # when switching from one demo to another.
    glViewport 0 0 $::g_WinWidth $::g_WinHeight

    glMatrixMode GL_MODELVIEW
    glLoadIdentity
    glTranslatef 0.0 0.0 -4.0
    glRotatef [expr -1.0 * $::g_fSpinY(1)] 1.0 0.0 0.0
    glRotatef [expr -1.0 * $::g_fSpinX(1)] 0.0 1.0 0.0

    if { $::g_isProgramLoaded } {
        # Identify the textures to use for the pixel shader...
        if { $::g_callSemantics eq "CG" } {
            cgGLSetTextureParameter $::g_CGparam_testTexture \
                                    [$::g_textureID($::g_testTextureInd) get 0]
            cgGLSetTextureParameter $::g_CGparam_checkerTexture \
                                    [$::g_textureID($::g_checkerTextureInd) get 0]
        }

        if { $::g_callSemantics eq "GL" } {
            # STAGE 0
            glActiveTextureARB GL_TEXTURE0_ARB
            glBindTexture GL_TEXTURE_2D [$::g_textureID($::g_testTextureInd) get 0]

            # STAGE 1
            glActiveTextureARB GL_TEXTURE1_ARB
            glBindTexture GL_TEXTURE_2D [$::g_textureID($::g_checkerTextureInd) get 0]
        }

        cgGLBindProgram $::g_CGprogram
        cgGLEnableProfile $::g_CGprofile

        if { $::g_callSemantics eq "CG" } {
            cgGLEnableTextureParameter $::g_CGparam_testTexture
            cgGLEnableTextureParameter $::g_CGparam_checkerTexture
        }
    }

    glInterleavedArrays GL_T2F_C3F_V3F 0 $::g_quadVertices
    glDrawArrays GL_QUADS 0 4

    if { $::g_isProgramLoaded } {
        if { $::g_callSemantics eq "CG" } {
            cgGLDisableTextureParameter $::g_CGparam_checkerTexture
            cgGLDisableTextureParameter $::g_CGparam_testTexture
        }

        cgGLDisableProfile $::g_CGprofile
    }

    $toglwin swapbuffers
}

proc CleanupShader {} {
    if { [info exists ::g_CGprogram] } {
        cgDestroyProgram $::g_CGprogram
    }
    if { [info exists ::g_CGcontext] } {
        cgDestroyContext $::g_CGcontext
    }
}

proc Cleanup {} {
    if { [info exists ::g_textureID($::g_testTextureInd)] } {
        glDeleteTextures 1 [$::g_textureID($::g_testTextureInd) get 0]
        $::g_textureID($::g_testTextureInd) delete
    }
    if { [info exists ::g_textureID($::g_checkerTextureInd)] } {
        glDeleteTextures 1 [$::g_textureID($::g_checkerTextureInd) get 0]
        $::g_textureID($::g_checkerTextureInd) delete
    }

    CleanupShader

    $::g_quadVertices delete

    foreach var [info globals g_*] {
        uplevel #0 unset $var
    }
}

proc ExitProg {} {
    exit
}

if { ! [tcl3dHaveCg] } {
    tk_messageBox -icon error -type ok -title "Error" \
                  -message "You do not have Cg installed."
    exit
}

frame .fr
pack .fr -expand 1 -fill both
togl .fr.toglwin -width $g_winWidth -height $g_winHeight \
                 -double true -depth true \
                 -createproc tclCreateFunc \
                 -reshapeproc tclReshapeFunc \
                 -displayproc tclDisplayFunc 
frame .fr.frActions 
label .fr.shaderinfo -bg white -anchor w
label .fr.info

grid .fr.toglwin    -row 0 -column 0 -sticky news
grid .fr.frActions  -row 1 -column 0 -sticky news
grid .fr.shaderinfo -row 2 -column 0 -sticky news
grid .fr.info       -row 3 -column 0 -sticky news
grid rowconfigure .fr 0 -weight 1
grid columnconfigure .fr 0 -weight 1

radiobutton .fr.frActions.bCG -text "Cg Semantics" \
            -variable g_callSemantics -value "CG" \
            -command "ToggleSemantics .fr.toglwin"
radiobutton .fr.frActions.bGL -text "OpenGL Semantics" \
            -variable g_callSemantics -value "GL" \
            -command "ToggleSemantics .fr.toglwin"
pack .fr.frActions.bCG .fr.frActions.bGL -side left -fill x

button .fr.frActions.viewFile -text "View shader file" \
                              -command ViewShaderFile \
                              -relief ridge
pack .fr.frActions.viewFile -side left -fill x

set appTitle "Tcl3D demo: CodeSampler's Multi-Texturing Using Cg"
wm title . "$appTitle"

# Watch For ESC Key And Quit Messages
wm protocol . WM_DELETE_WINDOW "ExitProg"
bind . <Key-Escape> "ExitProg"

bind .fr.toglwin <1>         "SetMouseInput %x %y"
bind .fr.toglwin <B1-Motion> "GetMouseInput %x %y"

PrintShaderInfo
PrintInfo [format "Running on %s with a %s (OpenGL %s, Tcl %s)" \
           $tcl_platform(os) [glGetString GL_RENDERER] \
           [glGetString GL_VERSION] [info patchlevel]]
